import waba.fx.*;     // Graphics
//import waba.io.*;
//import waba.sys.*;    // Convert
import waba.ui.*;
//import waba.util.*;   // Vector
import extra.ui.*;      // List

// Random

/**
 * @author <A HREF="mailto:weyer@kagi.com">Steve Weyer</A>
 * Copyright (c) 2001, S. Weyer. All right reserved.
 * @version waba 1.0
 * !!!d ?? Mar 2001 elim drawDots; tab; showCursor
 * !!!c 08 Mar 2001
 * !!!a 05 Mar 2001
 * <A HREF="http://www.kagi.com/weyer/#Crypto">Steve's other Crypto versions</A>
 * <A HREF="http://www.kagi.com/weyer/#Waba">Steve's other Waba apps</A>
 *
 * ! classSize: 22000 (List)
 * design notes
 * version a: CodeArea, CodeChar classes. problems: too much object heap
 * version b: elim CodeChar. problems: focus between CodeArea and other control obj
 *
 * ? different A-Z input
 * ? larger font
 * ? faster updates
**/

public class Crypto extends MainWindow {
  private static final boolean DEBUG = false;

  private static final int COMPHT  = 15;
  private static final int COMPWID = 38;
  private static final int COMPSEP = 2;

  final int NUMLETTERS = 26;
  final int LETTERHEIGHT = 9; // kbd row ht
  final int KBDLEFT = 4;      // allow for "space"
  private int kbdWid;
  final int codeTop = 20;     // top of code area
  final int codeWid = 10;     // width of edit field
  final int letHt = 14;       // guess char
  final int codeHt = letHt+10;  // +code char
  private int curEdit = -1;   // current 'edit field'
  private int rows;
  private int cols;

  final String ALPHA = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  private char[] lettersUsed = new char[NUMLETTERS];
  private char[] answer;
  private char[] code;
  private char[] guess;
  private Random rand;
  private List puzzleList;
  private Button hintButton;
  private Button clearButton;
  private Button enterButton;
  private Button encodeButton;
  private Graphics appG = null; // used for showCursor and as init flag

  private final String[] PUZZLEANSWERS = {
    "THIS IS CRYPTO. SELECT A PUZZLE THEN DECODE IT. HAVE FUN!--weyer@kagi.com",
    "", // NEW
    "Idealism increases in direct proportion to one's distance from the problem.",
    "Falsehood is so easy, truth so difficult. - George Eliot",
    "Sometime they'll give a war and nobody will come. - Carl Sandburg",
    "I had a boring office job. I cleaned the windows in the envelopes. - Rita Rudner",
    "Stupidity is doing the same thing over and over and expecting different results.",
    "When you enter the ocean, you enter the food chain...not necessarily at the top. -- Jacque-Yves Cousteau",
    "You know, I'm all for progress. It's change I object to. - Mark Twain",
    "The one serious conviction that a man should have is that nothing is to be taken too seriously. - Samuel Butler",
    "Do not worry about nobody knowing you; rather seek to be worth knowing. - Confucious",
    "No, try not. Do ... or do not. There is no try. - Yoda",
  };

  public Crypto() {
    rand = new Random();
    rows = (this.height-codeTop-COMPHT+2) / codeHt;
    cols = this.width / codeWid;
    kbdWid = this.width / NUMLETTERS;

    setTextCode(TrimUpString(PUZZLEANSWERS[0]), null, null); // assumes appG set later (paint flag)

    // bottom row of controls
    int objX = 0;
    int objWid = COMPWID+4;
    int objY = this.height - COMPHT;
    puzzleList = new List();
    puzzleList.add("About");
    puzzleList.add("New");
    for (int i=1; i<PUZZLEANSWERS.length-1; i++)
      puzzleList.add("puzzle" + i);
    puzzleList.setRect(objX, objY, objWid, COMPHT);
    //puzzleList.setSelectedIndex(0);
    add(puzzleList);

    objX += objWid;
    objWid = COMPWID-8;
    clearButton = new Button("Clear");
    clearButton.setRect(objX, objY, objWid, COMPHT);
    add(clearButton);

    objX += objWid+COMPSEP;
    enterButton = new Button("Enter");
    enterButton.setRect(objX, objY, objWid, COMPHT);
    add(enterButton);

    objX += objWid+COMPSEP;
    objWid = COMPWID;
    encodeButton = new Button("Encode");
    encodeButton.setRect(objX, objY, objWid, COMPHT);
    add(encodeButton);

    objX += objWid+COMPSEP;
    objWid = 16-COMPSEP;
    hintButton = new Button("?");
    hintButton.setRect(objX,objY,objWid,COMPHT);
    add(hintButton);
  }

  public void setTextCode(char[] ans, char[] cd, char[] key) {
    curEdit = -1;
    answer = ans;
    code = (cd == null) ? encodeString(answer, key) : cd;
    int len = code.length;
    guess = new char[len];

    if (appG == null) { // first time (display both answer & code)
      charFill(lettersUsed, ' ');
      char ch;
      for (int i=0; i<len; i++) {
        guess[i] = ch = answer[i];
        if (IsUpper(ch))
          lettersUsed[ch-'A'] = code[i];
      };
      //update via initial paint
    }
    else
      clear();
  }


  public void onPaint(Graphics g) {
    if (appG == null) { // first time (doesn't work in MRJ if in constructor)
      appG = createGraphics();  // for showCursor
      //appG.setClip(0, codeTop, this.width, this.height-codeTop-COMPHT);
    };

    //g.setColor(255,255,255);  // white
    //g.fillRect(0, 0, this.width, this.height-COMPHT);
    g.setColor(0,0,0);        // black

    int objX = KBDLEFT, objY = 0, i;
    // "keyboard" at top.  sp, A-Z
    // if proportional spacing, then enough room for W and final ?
    // but then more complex (see v1) and code underneath might have worse spacing
    for (i=0; i < NUMLETTERS; i++) {
      g.drawText(ALPHA.substring(i,i+1), objX, objY);
      if (lettersUsed[i] != ' ')
        g.drawText(new String(lettersUsed,i,1), objX, objY+LETTERHEIGHT);
      objX += kbdWid;
     };
     //g.drawText("?", objX, objY);

    char guessCh, codeCh = ' ';
    int len = guess.length;  // guess might be 0 length first time
    boolean create = (code.length == 0), dots = create; // New
    if (create && len == 0) // 1st time
      charFill(guess = new char[len = rows*cols], ' ');

    FontMetrics fm = getFontMetrics(defaultFont);
    // draw code area
    objY = codeTop;
    i = 0;
    for (int row=0; row < rows; row++) {
      objX = 0;
      for (int col=0; col < cols; col++) {
        if (i >= len)
          return;
        guessCh = guess[i];
        if (!create)
          dots = IsUpper(codeCh = code[i]);

        if (dots) {
          g.drawLine(objX,         objY+4,     objX,         objY+codeHt);  // |-
          g.drawLine(objX+codeWid, objY+4,     objX+codeWid, objY+codeHt);  //   -|
          g.drawLine(objX,         objY+letHt, objX+codeWid, objY+letHt);   //  --
        };
        // ?? String.valueOf(ch) vs. new String(ch[],i,1)
        if (guessCh != ' ') // draw char 'centered'
          g.drawText(String.valueOf(guessCh), objX+1+ (codeWid - fm.getCharWidth(guessCh))/2, objY+3);
        if (codeCh != ' ')
          g.drawText(String.valueOf(codeCh),  objX+1+ (codeWid - fm.getCharWidth(codeCh)) / 2, objY+letHt+1);
        if (curEdit == i)
          g.drawCursor(objX,objY+4,codeWid+1,letHt-4); // see showCursor
        i++;
        objX += codeWid;
      };
      objY += codeHt;
    };
  }

  public void showCursor() { // hide/show cursor. faster than complete repaint
    if (curEdit < 0)
      return;
    int col = curEdit % cols;
    int row = curEdit / cols;
    appG.drawCursor(col*codeWid, codeTop+4 + row*codeHt, codeWid+1, letHt-4);
  }

  public void onEvent(Event ev) {
    Object obj = ev.target;
    int evType = ev.type, len = guess.length;
    char guessCh = ' ';
    boolean create = (code.length == 0);
    if (evType == ControlEvent.PRESSED) {
      if (obj == hintButton)
        setChar('?');
      else if (obj == clearButton)
        clear();
      else if (obj == enterButton) {
        if (create) // after New
          setTextCode(null, TrimUpString(guess), null);
      }
      else if (obj == encodeButton) {
        if (create) // after New
          setTextCode(TrimUpString(guess), null, null);
      }
      else if (obj == puzzleList)
        setTextCode(TrimUpString(PUZZLEANSWERS[puzzleList.getSelectedIndex()]), null, null);
    }
    else if (obj == this && evType == PenEvent.PEN_DOWN) {
      int x = ((PenEvent) ev).x, col;
      int y = ((PenEvent) ev).y;
      if (y <= LETTERHEIGHT) {    // virtual kbd
        if (x >= KBDLEFT) {
          col = (x-KBDLEFT) / kbdWid;
          if (col < NUMLETTERS)
            guessCh = ALPHA.charAt(col);
        };
//System.out.println(guessCh);
        setChar(guessCh);
      }
      else if (y >= codeTop && y <= this.height-COMPHT) { // select edit field
        int row = (y-codeTop) / codeHt;
        col = x / codeWid;
        int i = col + (row * cols);
//System.out.println("col: " + col + "; row: " + row + "; i: " + i + "; len: " + len);
        if (i < len && (create || IsUpper(code[i]))) {
          showCursor();
          curEdit = i;
          showCursor();
        };
      };
    }
    else if (evType == KeyEvent.KEY_PRESS) {
      int key = ((KeyEvent) ev).key, inc = 0;
      guessCh = (char) key;
//System.out.println("ch: " + guessCh + "; key: " + key);
      if (key == IKeys.RIGHT || key == IKeys.TAB || key == 9)
        inc = 1;
      else if (key == IKeys.LEFT)
        inc = -1;
      else if (key == IKeys.UP)
        inc = -cols;
      else if (key == IKeys.DOWN)
        inc = cols;
      else if (key == IKeys.BACKSPACE) {
        setChar(' ');
        inc = -1;
      };
      if (inc != 0) { // move edit field
        showCursor();
        do {
          curEdit += inc;
          if (inc > 0)
            inc = 1;
          else if (inc < 0)
            inc = -1;
          if (curEdit < 0)
            curEdit = len-1;
          else if (curEdit >= len)
            curEdit = 0;
        }
        while (! (create || IsUpper(code[curEdit])));
        showCursor();
        return;
      }
      else if (key == IKeys.DELETE)
        guessCh = ' ';
      else if (IsLower(guessCh))
        guessCh += 'A'-'a';
      else if (guessCh == ' ' || guessCh == '?' || IsUpper(guessCh)) {
      }
      else if (!create)
        return;
      setChar(guessCh);
    }
  }

  public void clear() {
    charFill(lettersUsed, ' ');
    if (code.length == 0) // New
      charFill(guess, ' ');
    else {
      char ch;
      for(int i=0; i < guess.length; i++)
        guess[i] = (IsUpper(ch = code[i])) ? ' ' : ch;
    };
    repaint();
  }

  public void setChar(char guessCh) {
    if (curEdit < 0)
      return;
    if (code.length == 0) { // New
      guess[curEdit] = guessCh;
      repaint(); // ? overkill, but too messy/rare to update single char
    }
    else
      substitute(guessCh, code[curEdit]);
  }

  public void substitute(char guessCh, char codeCh) {
//System.out.println("sub: " + guessCh + "; code: " + codeCh);
    char oldCode;
    int gi = -1, oldi = charFind(lettersUsed, codeCh);
    // asking for hint?
    if (guessCh == '?') {
      if (answer == null) // none avail
        return;
      guessCh = answer[charFind(code, codeCh)];
      gi = guessCh-'A';
//System.out.println("answerCh: " + guessCh + "; gi: " + gi);
      oldCode = lettersUsed[gi];
      if (oldCode == codeCh)
        return;
      if (oldCode != ' ')
        substitute(' ', oldCode); // undo incorrect guess first
    }
    else if (guessCh == ' ') {
      if (oldi < 0) // nothing to erase
        return;
    }
    // UPPERCASE. guess already used?
    else if ((oldCode = lettersUsed[gi = guessCh-'A']) != ' ') {
      if (oldCode != codeCh) // already sub.
        Sound.beep();
      return;
    };

    // replace earlier guess for this code?
    if (oldi >= 0)
      lettersUsed[oldi] = ' ';

    if (gi >= 0)
      lettersUsed[gi] = codeCh;

    // do actual substitution
    for (int i=0; i < code.length; i++)
      if (code[i] == codeCh)
        guess[i] = guessCh;
    repaint();
  }

  public static int charFind(char[] chars, char ch) {
    for (int i=0; i < chars.length; i++)
      if (chars[i] == ch)
        return i;
    return -1;
  }

  public static void charFill(char[] chars, char ch) {
    for (int i=0; i < chars.length; i++)
      chars[i] = ch;
  }

  public char[] encodeString (char[] ans, char[] key) {
    if (key == null)
      key = generateKey();
    char ch;
    int len = ans.length;
    char[] enc = new char[len];
    for (int i = 0; i < len; i++)
      enc[i] = (IsUpper(ch = ans[i])) ? key[ch-'A'] : ch;
    return enc;
  }

  public char[] generateKey () {
    // create a random key
    int r;
    char ch;
    char[] k = ALPHA.toCharArray();
    for (int i = 0; i < NUMLETTERS; i++)
      if ((ch = k[i]) == ALPHA.charAt(i)) { 	// char still in original position?
        do
          r = rand.getRandom(NUMLETTERS);
        while (r == i);
	k[i] = k[r];
        k[r] = ch;
      };
    return k;
  }

  public static boolean IsWhiteSpace(char ch) {
    return ch == ' ' || ch == '\t' || ch == '\n';
  }
  public static boolean IsUpper(char ch) {
    return 'A' <= ch && ch <= 'Z';
  }
  public static boolean IsLower(char ch) {
    return 'a' <= ch && ch <= 'z';
  }

  // used usually before calling setTextCode
  public char[] TrimUpString(String str) {
    return TrimUpString(str.toCharArray());
  }
  public char[] TrimUpString (char[] tstr) {
    // get rid of leading, trailing and extra blanks in middle
    // but also convert cr,tab to space; lowercase to uppercase
    int ti = 0, len = tstr.length;
    char ch;
    boolean white = true; // ignore whitespace at beginning
    for (int si = 0; si < len; si++) {
      if (IsWhiteSpace(ch = tstr[si])) {
        if (! white) {
          tstr[ti++] = ' '; // convert/copy only first
          white = true;
        };
      }
      else {
        if (IsLower(ch))
          ch += 'A'-'a';            // make UPPERCASE
        tstr[ti++] = ch;
        white = false;
      };
    };
    if (white && ti > 0)  // remove final whitespace
      ti--;

    if (ti == len)
      return tstr;
    return new String(tstr,0,ti).toCharArray(); // ? Vm.copyArray
  }
}
