import waba.sys.Vm;

// Sean Luke <seanl@cs.umd.edu>  6/3/01

/** This is a simple Linear Congruential Generator which
    produces random numbers in the range [0,2^31), derived
    from ran0 in Numerical Recipies.  Note that ran0 isn't
    wonderfully random -- there are much better generators
    out there -- but it'll do the job, and it's fast and
    has low memory consumption.  */

public class Random
    {
    // Do NOT modify IA, IM, IQ, or IR
    // You are free to set ZERO_SEED_REPLACEMENT to any integer
    // greater than 0 (but NOT including 0!!)
    public static final int IA = 16807;
    public static final int IM = 2147483647;
    public static final int IQ = 127773;
    public static final int IR = 2836;
    public static final int ZERO_SEED_REPLACEMENT = 234123;  // whatever

    protected int seed;

    /** The only reasonable seeds are between 0 and 2^31 inclusive; if
        you're negative, it'll get the absolute value of it. */
    public Random(int _seed)
        {
        // strip out the minus-sign if any
        seed = (_seed << 1) >>> 1;
        // if seed is 0, it's invalid, we need another seed
        if (seed==0) seed = ZERO_SEED_REPLACEMENT;
        }

    /** Randomizes based on the current timestamp */
    public Random()
        {
        this(Vm.getTimeStamp());
        }

  /*  -- a useful debugging function
    static void printBits(int seed)
        {
        String s = "";
        for(int x=31;x>=0;x--)
           s = s + ((seed >>> x) % 2 == 1 ? "1" : "0");
        System.out.println(s);
        }
  */

    /** bits should be <= 31 */
    protected int next(final int bits)
        {
        int k = seed/IQ;
        seed = IA*(seed-k*IQ)-IR*k;
        if (seed < 0) seed += IM;
        return seed >> (31-bits);
        }

    /** Returns a floating-point value in the half-open range [0,1). */
    public float nextFloat()
        {
        return next(24) / ((float)(1 << 24));
        }

    /** Returns an integer in the range [0,n).
        n must be > 0, else -1 is returned. */
    public int nextInt(int n)
        {
        if (n<=0) return -1;
        int bits, val;
        do
            {
            bits = next(31);
            val = bits % n;
            }
        while(bits - val + (n-1) < 0);
        return val;
        }

    // SW. for compatibility with earlier Random
    public int getRandom(int n) {
      return nextInt(n);
    }

    }


